﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace LocalRefsInCSharp7 {

	class Vector {
		private int x;
		private int y;
	
		public ref int this[int component] {
			[System.Runtime.CompilerServices.MethodImpl(System.Runtime.CompilerServices.MethodImplOptions.AggressiveInlining)]
			get {
				if (component == 0) {
					return ref x;
				} if (component == 1) {
					return ref y;
				} else {
					throw new ArgumentOutOfRangeException(nameof(component), "Must be 0 for x, or 1 for y");
				}
			}
		}

		public void Print() {
			Console.WriteLine($"Vector(x={x}, y={y})");
		}
	}

	class VectorWithArray {
		private int[] items = new int[2];

		public ref int this[int component] {
			get {
				if (component < 0 || component > 1) { 
					throw new ArgumentOutOfRangeException(nameof(component), "Must be 0 or 1");
				}
				return ref items[component];
			}
		}

		public void Print() {
			Console.WriteLine($"VectorWithArray(items=[{items[0]}, {items[1]}])");
		}
	}

	class Program {
		static void Main(string[] args) {

			var v = new Vector();
			v[0] = 10;
			v[1] = 20;
			v.Print();
			int xComponent = v[0];
			Console.WriteLine($"local variable {nameof(xComponent)} == {xComponent}"); Console.WriteLine();

			for (int i = 0; i <= 1; i++) {
				v[i] = 0;
			}
			v.Print();

			#region Info about generated code
			// With [MethodImpl(MethodImplOptions.AggressiveInlining)] on getter of the indexer, the following code:
			//
			//   v[0] = 10;
			//   v[1] = 20;
			//
			// is jitted into (note that esi register contains the local variable v):
			//
			//		v[0] = 10;
			//   0142046C	lea eax,[esi + 4]			// lea = load effective address, i.e.: eax := esi + 4
			//   0142046F	mov dword ptr[eax],0Ah  
			//      v[1] = 20;
			//   01420475	lea eax,[esi + 8]			// lea = load effective address, i.e.: eax := esi + 8
			//   01420478	mov dword ptr[eax],14h
			//
			// Notice that all the ifs and throw got optimized out and we got code fully equivalent to direct field access.
			//
			// Works the same of x64:
			//   00007FF819BF04AF lea         rcx,[rsi+8]  
			//   00007FF819BF04B3 mov         dword ptr[rcx],0Ah  
			//   00007FF819BF04B9 lea         rcx,[rsi+0Ch]  
			//   00007FF819BF04BD mov         dword ptr[rcx],14h
			//
			// However for:
			//   for (int i = 0; i <= 1; i++) {
			//	     v[i] = 0;
			//   }
			// just for loop unrolling occurs, but no inlining occurs. Call to the getter body will be generated by x86 JIT.
			// The x64 JIT will also call the getter body + will not unroll the for loop.
			#endregion

			var va = new VectorWithArray();
			va[0] = 30;
			va[1] = 40;
			va.Print();
			int yComponent = va[1];
			Console.WriteLine($"local variable {nameof(yComponent)} == {yComponent}"); Console.WriteLine();

			int x = 5;
			int y = 20;

			Console.WriteLine($"x == {x}, y == {y}");
			Min(ref x, ref y) = 42;
			Console.WriteLine($"x == {x}, y == {y} after Min(ref x, ref y) = 42");
			Min(ref x, ref y) = 42;
			Console.WriteLine($"x == {x}, y == {y} after Min(ref x, ref y) = 42");
		}

		static ref int Min(ref int a, ref int b) {
			if (a <= b) {
				return ref a;
			} else {
				return ref b;
			}
		}

		static ref int m1(ref int a, ref int b) {
			// Error: A declaration of a by-reference variable must have an initializer
			// ref int x;

			// Error: Cannot initialize a by-reference variable with a value
			// ref int x = a;

			ref int x = ref a;

			// Error: By-value returns may only be used in methods that return by value
			// return x;

			return ref x;
		}

		static ref int m2(ref int a) {
			int local = 5;

			ref int x = ref local;

			// Error: Cannot return 'x' by reference because it was initialized
			//	      to a value that cannot be returned by reference
			// return ref x;

			ref int y = ref x;

			// Error: Cannot return 'x' by reference because it was initialized
			//        to a value that cannot be returned by reference
			// return ref y;

			// Error: Cannot return by reference a result of 'Program.m1(ref int, ref int)'
			//	      because the argument passed to parameter 'b' cannot be returned by reference
			// return ref m1(ref a, ref local);

			// Error: Cannot return by reference a result of 'Program.m1(ref int, ref int)'
			//	      because the argument passed to parameter 'b' cannot be returned by reference
			// return ref m1(ref a, ref y);

			return ref m1(ref a, ref a);
		}

	}
}
